from .. import BaseLock


class OpenLock(BaseLock):
    url_scheme = 'open'

    @classmethod
    def get_client(cls, **connection_args):
        return None

    def acquire(self):
        return True

    def release(self):
        return True



import unittest
import time

import redis

from nose.tools import raises
from nose.tools import eq_
from mock import Mock
from mock import patch


class TestRedisLock(unittest.TestCase):
    def _makeOne(self):
        from pylock import DEFAULT_EXPIRES, DEFAULT_TIMEOUT
        from pylock.backends.redis_lock import RedisLock

        # wrap the RedisLock backend in a context manager to be able to test it
        class Lock(RedisLock):
            def __init__(self, key, expires=DEFAULT_EXPIRES, timeout=DEFAULT_TIMEOUT, client=None):
                super(Lock, self).__init__(key, expires, timeout, client)

            def __enter__(self):
                self.acquire()

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.release()
        return Lock

    def _lockException(self):
        from pylock import LockTimeout
        return LockTimeout

    def test_redis_sub(self):
        mock_redis = Mock(spec=redis.Redis)

        def test_it():
            lock = self._makeOne()
            with lock('somekey', client=mock_redis):
                val = 2 + 4
        test_it()
        method_names = [x[0] for x in mock_redis.method_calls]
        eq_(method_names[0], 'setnx')
        eq_(method_names[1], 'expire')
        eq_(method_names[2], 'delete')
        eq_(len(method_names), 3)

    def test_nocontention(self):
        mock_redis = Mock(spec=redis.Redis)

        def test_it():
            lock = self._makeOne()
            with lock('somekey', client=mock_redis):
                val = 2 + 4
        test_it()
        method_names = [x[0] for x in mock_redis.method_calls]
        eq_(method_names[0], 'setnx')
        eq_(method_names[1], 'expire')
        eq_(method_names[2], 'delete')
        eq_(len(method_names), 3)

    def test_nocontention_and_no_lock_delete(self):
        mock_redis = Mock(spec=redis.Redis)
        mock_time = Mock()
        vals = [35, 0, 0, 0]
        mock_time.side_effect = lambda: vals.pop()

        @patch('time.time', mock_time)
        def test_it():
            lock = self._makeOne()
            with lock('somekey', expires=30, client=mock_redis):
                val = 2 + 4
        test_it()
        method_names = [x[0] for x in mock_redis.method_calls]
        eq_(method_names[0], 'setnx')
        assert 'delete' not in method_names

    def test_contention_and_someone_else_replacing_timeout(self):
        mock_redis = Mock(spec=redis.Redis)
        mock_time = Mock()
        vals = [35, 0, 0, 0]
        mock_time.side_effect = lambda: vals.pop()
        mock_redis.get.return_value = False
        mock_redis.setnx.return_value = False
        timeout = self._lockException()

        @raises(timeout)
        @patch('time.time', mock_time)
        @patch('time.sleep', Mock())
        def test_it():
            lock = self._makeOne()
            with lock('somekey', expires=30, timeout=0, client=mock_redis):  # pragma: nocover
                val = 2 + 4
        test_it()
        method_names = [x[0] for x in mock_redis.method_calls]
        eq_(method_names[0], 'setnx')
        eq_(len(method_names), 2)

    def test_contention(self):
        mock_redis = Mock(spec=redis.Redis)
        mock_redis.get.return_value = 150
        mock_redis.getset.return_value = 150
        mock_redis.setnx.return_value = False

        def test_it():
            lock = self._makeOne()
            with lock('somekey', client=mock_redis):
                val = 2 + 4
        test_it()
        eq_(len(mock_redis.method_calls), 5)
        setnx, get = mock_redis.method_calls[:2]
        eq_(setnx[1][0], 'somekey')
        eq_(get[1][0], 'somekey')

    def test_timeout_current_val_is_newer(self):
        mock_redis = Mock(spec=redis.Redis)
        mock_redis.setnx.return_value = False
        mock_redis.get.return_value = time.time() + 200
        timeout = self._lockException()
        mock_time = Mock()
        mock_time.return_value = 0

        array = []

        @raises(timeout)
        @patch('time.sleep', mock_time)
        def test_it():
            lock = self._makeOne()
            with lock('somekey', timeout=1, client=mock_redis):
                array.append(4)  # pragma: nocover
        test_it()
        eq_(len(array), 0)

    def test_timeout_no_current_value_and_mock_redis(self):
        mock_redis = Mock(spec=redis.Redis)
        mock_redis.setnx.return_value = False
        mock_redis.get.return_value = 150
        mock_redis.getset.return_value = False
        timeout = self._lockException()
        mock_time = Mock()
        mock_time.return_value = 0

        array = []

        @raises(timeout)
        @patch('time.sleep', mock_time)
        def test_it():
            lock = self._makeOne()
            with lock('somekey', timeout=1, client=mock_redis):
                array.append(4)  # pragma: nocover
        test_it()
        eq_(len(array), 0)

    def test_get_client(self):
        from pylock.backends.redis_lock import RedisLock
        db = 4
        host = '33.33.33.10'
        port = '1234'
        password = 'cookies'
        with patch('pylock.backends.redis_lock.StrictRedis') as mock_redis:
            RedisLock.get_client(db=db, host=host, port=port, password=password)
            mock_redis.assert_called_once_with(host, port, db, password)

    def test_get_client_default_connection_values(self):
        from pylock.backends.redis_lock import RedisLock
        db = None
        host = None
        port = None
        password = None
        with patch('pylock.backends.redis_lock.StrictRedis') as mock_redis:
            RedisLock.get_client(db=db, host=host, port=port, password=password)
            mock_redis.assert_called_once_with('localhost', 6379, 0, None)



class LockTimeout(BaseException):
    """Raised in the event a timeout occurs while waiting for a lock"""


class BaseLock(object):
    def __init__(self, key, expires, timeout, client):
        """
        :param  client:      The the backend client instance to use.

        """
        self.key = key
        self.expires = expires
        self.timeout = timeout
        self.client = client

    @classmethod
    def get_client(cls, **connection_args):
        raise NotImplementedError

    def acquire(self):
        raise NotImplementedError

    def release(self):
        raise NotImplementedError



import time

from redis import StrictRedis

from .. import BaseLock, LockTimeout


class RedisLock(BaseLock):
    url_scheme = 'redis'

    @classmethod
    def get_client(cls, **connection_args):
        host = connection_args.get('host') or 'localhost'
        port = connection_args.get('port') or 6379
        password = connection_args.get('password')
        db = connection_args.get('db') or 0
        return StrictRedis(host, port, db, password)

    def __init__(self, key, expires, timeout, client):
        super(RedisLock, self).__init__(key, expires, timeout, client)
        self.start_time = time.time()

    def acquire(self):
        redis = self.client
        timeout = self.timeout
        while timeout >= 0:
            expires = time.time() + self.expires + 1

            if redis.setnx(self.key, expires):
                # We gained the lock; enter critical section
                self.start_time = time.time()
                redis.expire(self.key, int(self.expires))
                return

            current_value = redis.get(self.key)

            # We found an expired lock
            if current_value and float(current_value) < time.time():
                # Nobody raced us to replacing it
                if redis.getset(self.key, expires) == current_value:
                    self.start_time = time.time()
                    redis.expire(self.key, int(self.expires))
                    return

            timeout -= 1
            if timeout >= 0:
                time.sleep(1)
        raise LockTimeout("Timeout while waiting for lock")

    def release(self):
        redis = self.client
        # Only delete the key if we completed within the lock expiration,
        # otherwise, another lock might have been established
        if time.time() - self.start_time < self.expires:
            redis.delete(self.key)



from importlib import import_module
import logging
from six.moves.urllib import parse

from .backends import LockTimeout  # noqa

DEFAULT_TIMEOUT = 60
DEFAULT_EXPIRES = 10
KEY_PREFIX = 'pylock:'
DEFAULT_BACKEND = {
    'class': 'pylock.backends.redis_lock.RedisLock',
    'connection': 'redis://'
}


logger = logging.getLogger('pylock')


class Lock(object):
    """
    Distributed locking.

    Usage::

        with Lock('my_lock'):
            print "Critical section"

    :param  key:        The key against which the lock will be held.
    :param  expires:    We consider any existing lock older than
                        ``expires`` seconds to be invalid in order to
                        detect crashed clients. This value must be higher
                        than it takes the critical section to execute.
    :param  timeout:    If another client has already obtained the lock,
                        sleep for a maximum of ``timeout`` seconds before
                        giving up. A value of 0 means no wait (give up
                        right away).

    """

    def __init__(self, key, expires=None, timeout=None, backend_class_path=None, backend_connection=None):
        if expires is None:
            expires = DEFAULT_EXPIRES
        if timeout is None:
            timeout = DEFAULT_TIMEOUT
        if backend_class_path is None:
            backend_class_path = DEFAULT_BACKEND['class']
        if backend_connection is None:
            backend_connection = DEFAULT_BACKEND['connection']
        # Load backend class
        backend_class = get_backend_class(backend_class_path)
        logger.info("Using {0} lock backend".format(backend_class.__name__))
        key = "{0}{1}".format(KEY_PREFIX, key)
        connection_info = parse_url(backend_connection, url_scheme=backend_class.url_scheme)
        client = backend_class.get_client(**connection_info)
        self._lock = backend_class(key, expires, timeout, client)

    def __enter__(self):
        self._lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lock.release()


class ImproperlyConfigured(Exception):
    pass


def get_backend_class(import_path):
    try:
        dot = import_path.rindex('.')
    except ValueError:
        raise ImproperlyConfigured("%s isn't a pylock backend module." % import_path)
    module, classname = import_path[:dot], import_path[dot + 1:]
    try:
        mod = import_module(module)
    except ImportError as e:
        raise ImproperlyConfigured('Error importing pylock backend module %s: "%s"' % (module, e))
    try:
        return getattr(mod, classname)
    except AttributeError:
        raise ImproperlyConfigured('Pylock backend module "%s" does not define a "%s" class.' % (module, classname))


def parse_url(url, url_scheme):
    """Parses a distributed lock backend URL."""
    # Register extra schemes in URLs.
    parse.uses_netloc.append(url_scheme)

    url = parse.urlparse(url)

    # Remove query strings.
    path = url.path[1:]
    path = path.split('?', 2)[0]

    # Update with environment configuration.
    connection_info = {
        'db': path,
        'user': url.username,
        'password': url.password,
        'host': url.hostname,
        'port': url.port
    }

    return connection_info



import os

from setuptools import setup, find_packages

__version__ = '0.3'

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

setup(
    name='pylock',
    version=__version__,
    description='Python Distributed Lock',
    long_description=README,
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
    ],
    keywords='lock redis',
    author="Nathan Duthoit",
    author_email="nathan@waveapps.com",
    url="http://github.com/waveaccounting/pylock",
    license="BSD",
    packages=find_packages(),
    test_suite="pylock.tests",
    include_package_data=True,
    zip_safe=False,
    tests_require=['Mock>=2.0.0', 'nose'],
    install_requires=[
        "redis>=2.6.1",
    ]
)



import unittest

from nose.tools import eq_
from mock import Mock

from pylock.backends.open_lock import OpenLock
from pylock.backends.redis_lock import RedisLock


class TestLock(unittest.TestCase):
    def setUp(self):
        import pylock
        self.current_settings = (pylock.DEFAULT_BACKEND, pylock.DEFAULT_EXPIRES, pylock.DEFAULT_TIMEOUT, pylock.KEY_PREFIX)

    def tearDown(self):
        import pylock
        (pylock.DEFAULT_BACKEND, pylock.DEFAULT_EXPIRES, pylock.DEFAULT_TIMEOUT, pylock.KEY_PREFIX) = self.current_settings

    def test_open_backend_is_properly_detected(self):
        import pylock
        pylock.DEFAULT_BACKEND = {'class': 'pylock.backends.open_lock.OpenLock', 'connection': 'open://'}
        lock = pylock.Lock('somekey')
        self.assertEqual(lock._lock.__class__, OpenLock)

    def test_redis_backend_is_properly_detected(self):
        import pylock
        pylock.DEFAULT_BACKEND = {'class': 'pylock.backends.redis_lock.RedisLock', 'connection': 'redis://'}
        lock = pylock.Lock('somekey')
        self.assertEqual(lock._lock.__class__, RedisLock)

    def test_default_values_are_used(self):
        import pylock
        pylock.DEFAULT_BACKEND = {'class': 'pylock.backends.open_lock.OpenLock', 'connection': 'open://'}
        pylock.DEFAULT_EXPIRES = 999
        pylock.DEFAULT_TIMEOUT = 888
        pylock.KEY_PREFIX = 'cookies-are-us-'
        lock = pylock.Lock('somekey')
        expected_key = "cookies-are-us-somekey"
        self.assertEqual(lock._lock.key, expected_key)
        self.assertEqual(lock._lock.expires, 999)
        self.assertEqual(lock._lock.timeout, 888)

    def test_context_manager_calls_backend_methods(self):
        import pylock
        lock = pylock.Lock('somekey')
        lock._lock = Mock()

        def test_it():
            with lock:
                val = 2 + 4

        test_it()

        # The context manager should call the lock backend's acquire method on entry and the release method
        # on exit
        method_names = [x[0] for x in lock._lock.method_calls]
        eq_(method_names[0], 'acquire')
        eq_(method_names[1], 'release')
        eq_(len(method_names), 2)

    def test_lock_is_released_on_exception(self):
        import pylock
        lock = pylock.Lock('somekey')
        lock._lock = Mock()

        def test_it():
            with lock:
                raise Exception()

        try:
            test_it()
        except Exception:
            pass

        # The context manager should call the lock backend's acquire method on entry and the release method
        # when the exception is raised
        method_names = [x[0] for x in lock._lock.method_calls]
        eq_(method_names[0], 'acquire')
        eq_(method_names[1], 'release')
        eq_(len(method_names), 2)
